@@ -1,15 +1,15 @@
 /* -----------------------------------------------------------------------------
  * Programmer(s): David J. Gardner @ LLNL
  * -----------------------------------------------------------------------------
- * LLNS Copyright Start
- * Copyright (c) 2014, Lawrence Livermore National Security
- * This work was performed under the auspices of the U.S. Department
- * of Energy by Lawrence Livermore National Laboratory in part under
- * Contract W-7405-Eng-48 and in part under Contract DE-AC52-07NA27344.
- * Produced at the Lawrence Livermore National Laboratory.
+ * SUNDIALS Copyright Start
+ * Copyright (c) 2002-2020, Lawrence Livermore National Security
+ * and Southern Methodist University.
  * All rights reserved.
- * For details, see the LICENSE file.
- * LLNS Copyright End
+ *
+ * See the top-level LICENSE and NOTICE files for details.
+ *
+ * SPDX-License-Identifier: BSD-3-Clause
+ * SUNDIALS Copyright End
  * -----------------------------------------------------------------------------
  * This the implementation file for the CVODES nonlinear solver interface.
  * ---------------------------------------------------------------------------*/
@@ -36,11 +36,9 @@ static int cvNlsResidualSensSim(N_Vector ycorSim, N_Vector resSim,
 static int cvNlsFPFunctionSensSim(N_Vector ycorSim, N_Vector resSim,
                                   void* cvode_mem);
 
-static int cvNlsLSetupSensSim(N_Vector ycorSim, N_Vector resSim,
-                              booleantype jbad, booleantype* jcur,
-                              void* cvode_mem);
-static int cvNlsLSolveSensSim(N_Vector ycorSim, N_Vector deltaSim,
+static int cvNlsLSetupSensSim(booleantype jbad, booleantype* jcur,
                               void* cvode_mem);
+static int cvNlsLSolveSensSim(N_Vector deltaSim, void* cvode_mem);
 static int cvNlsConvTestSensSim(SUNNonlinearSolver NLS,
                                 N_Vector ycorSim, N_Vector delSim,
                                 realtype tol, N_Vector ewtSim, void* cvode_mem);
@@ -72,9 +70,7 @@ int CVodeSetNonlinearSolverSensSim(void *cvode_mem, SUNNonlinearSolver NLS)
 
   /* check for required nonlinear solver functions */
   if ( NLS->ops->gettype    == NULL ||
-       NLS->ops->initialize == NULL ||
        NLS->ops->solve      == NULL ||
-       NLS->ops->free       == NULL ||
        NLS->ops->setsysfn   == NULL ) {
     cvProcessError(cv_mem, CV_ILL_INPUT, "CVODES",
                    "CVodeSetNonlinearSolverSensSim",
@@ -99,7 +95,8 @@ int CVodeSetNonlinearSolverSensSim(void *cvode_mem, SUNNonlinearSolver NLS)
   }
 
   /* free any existing nonlinear solver */
-  if (cv_mem->NLSsim) retval = SUNNonlinSolFree(cv_mem->NLSsim);
+  if ((cv_mem->NLSsim != NULL) && (cv_mem->ownNLSsim))
+    retval = SUNNonlinSolFree(cv_mem->NLSsim);
 
   /* set SUNNonlinearSolver pointer */
   cv_mem->NLSsim = NLS;
@@ -128,7 +125,8 @@ int CVodeSetNonlinearSolverSensSim(void *cvode_mem, SUNNonlinearSolver NLS)
   }
 
   /* set convergence test function */
-  retval = SUNNonlinSolSetConvTestFn(cv_mem->NLSsim, cvNlsConvTestSensSim);
+  retval = SUNNonlinSolSetConvTestFn(cv_mem->NLSsim, cvNlsConvTestSensSim,
+                                     cvode_mem);
   if (retval != CV_SUCCESS) {
     cvProcessError(cv_mem, CV_ILL_INPUT, "CVODES",
                    "CVodeSetNonlinearSolverSensSim",
@@ -148,8 +146,8 @@ int CVodeSetNonlinearSolverSensSim(void *cvode_mem, SUNNonlinearSolver NLS)
   /* create vector wrappers if necessary */
   if (cv_mem->simMallocDone == SUNFALSE) {
 
-    cv_mem->ycor0Sim = N_VNewEmpty_SensWrapper(cv_mem->cv_Ns+1);
-    if (cv_mem->ycor0Sim == NULL) {
+    cv_mem->zn0Sim = N_VNewEmpty_SensWrapper(cv_mem->cv_Ns+1);
+    if (cv_mem->zn0Sim == NULL) {
       cvProcessError(cv_mem, CV_MEM_FAIL, "CVODES",
                      "CVodeSetNonlinearSolverSensSim", MSGCV_MEM_FAIL);
       return(CV_MEM_FAIL);
@@ -157,7 +155,7 @@ int CVodeSetNonlinearSolverSensSim(void *cvode_mem, SUNNonlinearSolver NLS)
 
     cv_mem->ycorSim = N_VNewEmpty_SensWrapper(cv_mem->cv_Ns+1);
     if (cv_mem->ycorSim == NULL) {
-      N_VDestroy(cv_mem->ycor0Sim);
+      N_VDestroy(cv_mem->zn0Sim);
       cvProcessError(cv_mem, CV_MEM_FAIL, "CVODES",
                      "CVodeSetNonlinearSolverSensSim", MSGCV_MEM_FAIL);
       return(CV_MEM_FAIL);
@@ -165,7 +163,7 @@ int CVodeSetNonlinearSolverSensSim(void *cvode_mem, SUNNonlinearSolver NLS)
 
     cv_mem->ewtSim = N_VNewEmpty_SensWrapper(cv_mem->cv_Ns+1);
     if (cv_mem->ewtSim == NULL) {
-      N_VDestroy(cv_mem->ycor0Sim);
+      N_VDestroy(cv_mem->zn0Sim);
       N_VDestroy(cv_mem->ycorSim);
       cvProcessError(cv_mem, CV_MEM_FAIL, "CVODES",
                      "CVodeSetNonlinearSolverSensSim", MSGCV_MEM_FAIL);
@@ -176,16 +174,19 @@ int CVodeSetNonlinearSolverSensSim(void *cvode_mem, SUNNonlinearSolver NLS)
   }
 
   /* attach vectors to vector wrappers */
-  NV_VEC_SW(cv_mem->ycor0Sim, 0) = cv_mem->cv_tempv;
-  NV_VEC_SW(cv_mem->ycorSim,  0) = cv_mem->cv_acor;
-  NV_VEC_SW(cv_mem->ewtSim,   0) = cv_mem->cv_ewt;
+  NV_VEC_SW(cv_mem->zn0Sim,  0) = cv_mem->cv_zn[0];
+  NV_VEC_SW(cv_mem->ycorSim, 0) = cv_mem->cv_acor;
+  NV_VEC_SW(cv_mem->ewtSim,  0) = cv_mem->cv_ewt;
 
   for (is=0; is < cv_mem->cv_Ns; is++) {
-    NV_VEC_SW(cv_mem->ycor0Sim, is+1) = cv_mem->cv_tempvS[is];
-    NV_VEC_SW(cv_mem->ycorSim,  is+1) = cv_mem->cv_acorS[is];
-    NV_VEC_SW(cv_mem->ewtSim,   is+1) = cv_mem->cv_ewtS[is];
+    NV_VEC_SW(cv_mem->zn0Sim,  is+1) = cv_mem->cv_znS[0][is];
+    NV_VEC_SW(cv_mem->ycorSim, is+1) = cv_mem->cv_acorS[is];
+    NV_VEC_SW(cv_mem->ewtSim,  is+1) = cv_mem->cv_ewtS[is];
   }
 
+  /* Reset the acnrmcur flag to SUNFALSE */
+  cv_mem->cv_acnrmcur = SUNFALSE;
+
   return(CV_SUCCESS);
 }
 
@@ -236,8 +237,7 @@ int cvNlsInitSensSim(CVodeMem cvode_mem)
 }
 
 
-static int cvNlsLSetupSensSim(N_Vector ycorSim, N_Vector resSim,
-                              booleantype jbad, booleantype* jcur,
+static int cvNlsLSetupSensSim(booleantype jbad, booleantype* jcur,
                               void* cvode_mem)
 {
   CVodeMem cv_mem;
@@ -278,7 +278,7 @@ static int cvNlsLSetupSensSim(N_Vector ycorSim, N_Vector resSim,
 }
 
 
-static int cvNlsLSolveSensSim(N_Vector ycorSim, N_Vector deltaSim, void* cvode_mem)
+static int cvNlsLSolveSensSim(N_Vector deltaSim, void* cvode_mem)
 {
   CVodeMem cv_mem;
   int retval, is;
@@ -379,6 +379,7 @@ static int cvNlsConvTestSensSim(SUNNonlinearSolver NLS,
       cv_mem->cv_acnrm = (cv_mem->cv_errconS) ?
         N_VWrmsNorm(ycorSim, ewtSim) : N_VWrmsNorm(ycor, ewt);
     }
+    cv_mem->cv_acnrmcur = SUNTRUE;
     return(CV_SUCCESS);
   }
 
