@@ -1,19 +1,15 @@
 /* -----------------------------------------------------------------------------
  * Programmer(s): Daniel R. Reynolds @ SMU
  * -----------------------------------------------------------------------------
- * LLNS/SMU Copyright Start
- * Copyright (c) 2018, Southern Methodist University and
- * Lawrence Livermore National Security
- *
- * This work was performed under the auspices of the U.S. Department
- * of Energy by Southern Methodist University and Lawrence Livermore
- * National Laboratory under Contract DE-AC52-07NA27344.
- * Produced at Southern Methodist University and the Lawrence
- * Livermore National Laboratory.
- *
+ * SUNDIALS Copyright Start
+ * Copyright (c) 2002-2020, Lawrence Livermore National Security
+ * and Southern Methodist University.
  * All rights reserved.
- * For details, see the LICENSE file.
- * LLNS/SMU Copyright End
+ *
+ * See the top-level LICENSE and NOTICE files for details.
+ *
+ * SPDX-License-Identifier: BSD-3-Clause
+ * SUNDIALS Copyright End
  * -----------------------------------------------------------------------------
  * This is the implementation file for the SUNNonlinearSolver module
  * implementation of the Anderson-accelerated Fixed-Point method.
@@ -48,7 +44,6 @@ static void FreeContent(SUNNonlinearSolver NLS);
 SUNNonlinearSolver SUNNonlinSol_FixedPoint(N_Vector y, int m)
 {
   SUNNonlinearSolver NLS;
-  SUNNonlinearSolver_Ops ops;
   SUNNonlinearSolverContent_FixedPoint content;
   int retval;
 
@@ -65,61 +60,47 @@ SUNNonlinearSolver SUNNonlinSol_FixedPoint(N_Vector y, int m)
 
   /* Create nonlinear linear solver */
   NLS = NULL;
-  NLS = (SUNNonlinearSolver) malloc(sizeof *NLS);
+  NLS = SUNNonlinSolNewEmpty();
   if (NLS == NULL) return(NULL);
 
-  /* Create nonlinear solver operations structure */
-  ops = NULL;
-  ops = (SUNNonlinearSolver_Ops) malloc(sizeof *ops);
-  if (ops == NULL) { free(NLS); return(NULL); }
+  /* Attach operations */
+  NLS->ops->gettype         = SUNNonlinSolGetType_FixedPoint;
+  NLS->ops->initialize      = SUNNonlinSolInitialize_FixedPoint;
+  NLS->ops->solve           = SUNNonlinSolSolve_FixedPoint;
+  NLS->ops->free            = SUNNonlinSolFree_FixedPoint;
+  NLS->ops->setsysfn        = SUNNonlinSolSetSysFn_FixedPoint;
+  NLS->ops->setctestfn      = SUNNonlinSolSetConvTestFn_FixedPoint;
+  NLS->ops->setmaxiters     = SUNNonlinSolSetMaxIters_FixedPoint;
+  NLS->ops->getnumiters     = SUNNonlinSolGetNumIters_FixedPoint;
+  NLS->ops->getcuriter      = SUNNonlinSolGetCurIter_FixedPoint;
+  NLS->ops->getnumconvfails = SUNNonlinSolGetNumConvFails_FixedPoint;
 
   /* Create nonlinear solver content structure */
   content = NULL;
   content = (SUNNonlinearSolverContent_FixedPoint) malloc(sizeof *content);
-  if (content == NULL) { free(ops); free(NLS); return(NULL); }
-
-  /* Attach content and ops */
-  NLS->content = content;
-  NLS->ops     = ops;
-
-  /* Attach operations */
-  ops->gettype         = SUNNonlinSolGetType_FixedPoint;
-  ops->initialize      = SUNNonlinSolInitialize_FixedPoint;
-  ops->setup           = NULL;  /* no setup needed */
-  ops->solve           = SUNNonlinSolSolve_FixedPoint;
-  ops->free            = SUNNonlinSolFree_FixedPoint;
-  ops->setsysfn        = SUNNonlinSolSetSysFn_FixedPoint;
-  ops->setlsetupfn     = NULL;  /* no lsetup needed */
-  ops->setlsolvefn     = NULL;  /* no lsolve needed */
-  ops->setctestfn      = SUNNonlinSolSetConvTestFn_FixedPoint;
-  ops->setmaxiters     = SUNNonlinSolSetMaxIters_FixedPoint;
-  ops->getnumiters     = SUNNonlinSolGetNumIters_FixedPoint;
-  ops->getcuriter      = SUNNonlinSolGetCurIter_FixedPoint;
-  ops->getnumconvfails = SUNNonlinSolGetNumConvFails_FixedPoint;
+  if (content == NULL) { SUNNonlinSolFree(NLS); return(NULL); }
 
   /* Initialize all components of content to 0/NULL */
   memset(content, 0, sizeof(struct _SUNNonlinearSolverContent_FixedPoint));
 
+  /* Attach content */
+  NLS->content = content;
+
   /* Fill general content */
   content->Sys        = NULL;
   content->CTest      = NULL;
   content->m          = m;
+  content->damping    = SUNFALSE;
+  content->beta       = ONE;
   content->curiter    = 0;
   content->maxiters   = 3;
   content->niters     = 0;
   content->nconvfails = 0;
+  content->ctest_data = NULL;
 
   /* Fill allocatable content */
   retval = AllocateContent(NLS, y);
-
-  if (retval != SUN_NLS_SUCCESS) {
-    NLS->content = NULL;
-    NLS->ops = NULL;
-    free(content);
-    free(ops);
-    free(NLS);
-    return(NULL);
-  }
+  if (retval != SUN_NLS_SUCCESS) { SUNNonlinSolFree(NLS); return(NULL); }
 
   return(NLS);
 }
@@ -193,7 +174,7 @@ int SUNNonlinSolInitialize_FixedPoint(SUNNonlinearSolver NLS)
   by the Sys function provided to the nonlinear solver.
   ---------------------------------------------------------------------------*/
 int SUNNonlinSolSolve_FixedPoint(SUNNonlinearSolver NLS, N_Vector y0,
-                                 N_Vector y, N_Vector w, realtype tol,
+                                 N_Vector ycor, N_Vector w, realtype tol,
                                  booleantype callSetup, void* mem)
 {
   /* local variables */
@@ -201,7 +182,19 @@ int SUNNonlinSolSolve_FixedPoint(SUNNonlinearSolver NLS, N_Vector y0,
   N_Vector yprev, gy, delta;
 
   /* check that the inputs are non-null */
-  if ( (NLS == NULL) || (y0 == NULL) || (y == NULL) || (w == NULL) || (mem == NULL) )
+  if ( (NLS  == NULL) ||
+       (y0   == NULL) ||
+       (ycor == NULL) ||
+       (w    == NULL) ||
+       (mem  == NULL) )
+    return(SUN_NLS_MEM_NULL);
+
+  /* check that all required function pointers have been set */
+  if ( (FP_CONTENT(NLS)->Sys == NULL) || (FP_CONTENT(NLS)->CTest == NULL) )
+    return(SUN_NLS_MEM_NULL);
+
+  /* check that all required function pointers have been set */
+  if ( (FP_CONTENT(NLS)->Sys == NULL) || (FP_CONTENT(NLS)->CTest == NULL) )
     return(SUN_NLS_MEM_NULL);
 
   /* set local shortcut variables */
@@ -209,9 +202,6 @@ int SUNNonlinSolSolve_FixedPoint(SUNNonlinearSolver NLS, N_Vector y0,
   gy    = FP_CONTENT(NLS)->gy;
   delta = FP_CONTENT(NLS)->delta;
 
-  /* load prediction into y */
-  N_VScale(ONE, y0, y);
-
   /* Looping point for attempts at solution of the nonlinear system:
        Evaluate fixed-point function (store in gy).
        Performs the accelerated fixed-point iteration.
@@ -221,27 +211,28 @@ int SUNNonlinSolSolve_FixedPoint(SUNNonlinearSolver NLS, N_Vector y0,
        FP_CONTENT(NLS)->curiter++ ) {
 
     /* update previous solution guess */
-    N_VScale(ONE, y, yprev);
+    N_VScale(ONE, ycor, yprev);
 
     /* compute fixed-point iteration function, store in gy */
-    retval = FP_CONTENT(NLS)->Sys(y, gy, mem);
+    retval = FP_CONTENT(NLS)->Sys(ycor, gy, mem);
     if (retval != SUN_NLS_SUCCESS) break;
 
     /* perform fixed point update, based on choice of acceleration or not */
     if (FP_CONTENT(NLS)->m == 0) {    /* basic fixed-point solver */
-      N_VScale(ONE, gy, y);
+      N_VScale(ONE, gy, ycor);
     } else {                          /* Anderson-accelerated solver */
-      retval = AndersonAccelerate(NLS, gy, y, yprev, FP_CONTENT(NLS)->curiter);
+      retval = AndersonAccelerate(NLS, gy, ycor, yprev, FP_CONTENT(NLS)->curiter);
     }
 
     /* increment nonlinear solver iteration counter */
     FP_CONTENT(NLS)->niters++;
 
     /* compute change in solution, and call the convergence test function */
-    N_VLinearSum(ONE, y, -ONE, yprev, delta);
+    N_VLinearSum(ONE, ycor, -ONE, yprev, delta);
 
     /* test for convergence */
-    retval = FP_CONTENT(NLS)->CTest(NLS, y, delta, tol, w, mem);
+    retval = FP_CONTENT(NLS)->CTest(NLS, ycor, delta, tol, w,
+                                    FP_CONTENT(NLS)->ctest_data);
 
     /* return if successful */
     if (retval == SUN_NLS_SUCCESS)  return(SUN_NLS_SUCCESS);
@@ -305,7 +296,9 @@ int SUNNonlinSolSetSysFn_FixedPoint(SUNNonlinearSolver NLS, SUNNonlinSolSysFn Sy
   return(SUN_NLS_SUCCESS);
 }
 
-int SUNNonlinSolSetConvTestFn_FixedPoint(SUNNonlinearSolver NLS, SUNNonlinSolConvTestFn CTestFn)
+int SUNNonlinSolSetConvTestFn_FixedPoint(SUNNonlinearSolver NLS,
+                                         SUNNonlinSolConvTestFn CTestFn,
+                                         void* ctest_data)
 {
   /* check that the nonlinear solver is non-null */
   if (NLS == NULL)
@@ -316,6 +309,10 @@ int SUNNonlinSolSetConvTestFn_FixedPoint(SUNNonlinearSolver NLS, SUNNonlinSolCon
     return(SUN_NLS_ILL_INPUT);
 
   FP_CONTENT(NLS)->CTest = CTestFn;
+
+  /* attach convergence test data */
+  FP_CONTENT(NLS)->ctest_data = ctest_data;
+
   return(SUN_NLS_SUCCESS);
 }
 
@@ -333,6 +330,29 @@ int SUNNonlinSolSetMaxIters_FixedPoint(SUNNonlinearSolver NLS, int maxiters)
   return(SUN_NLS_SUCCESS);
 }
 
+int SUNNonlinSolSetDamping_FixedPoint(SUNNonlinearSolver NLS, realtype beta)
+{
+  /* check that the nonlinear solver is non-null */
+  if (NLS == NULL)
+    return(SUN_NLS_MEM_NULL);
+
+  /* check that beta is a vaild */
+  if (beta <= ZERO)
+    return(SUN_NLS_ILL_INPUT);
+
+  if (beta < ONE) {
+    /* enable damping */
+    FP_CONTENT(NLS)->beta    = beta;
+    FP_CONTENT(NLS)->damping = SUNTRUE;
+  } else {
+    /* disable damping */
+    FP_CONTENT(NLS)->beta    = ONE;
+    FP_CONTENT(NLS)->damping = SUNFALSE;
+  }
+
+  return(SUN_NLS_SUCCESS);
+}
+
 
 /*==============================================================================
   Get functions
@@ -407,9 +427,10 @@ static int AndersonAccelerate(SUNNonlinearSolver NLS, N_Vector gval,
                               N_Vector x, N_Vector xold, int iter)
 {
   /* local variables */
-  int       nvec, retval, i_pt, i, j, lAA, maa, *ipt_map;
-  realtype  a, b, rtemp, c, s, *cvals, *R, *gamma;
-  N_Vector  fv, vtemp, gold, fold, *df, *dg, *Q, *Xvecs;
+  int         nvec, retval, i_pt, i, j, lAA, maa, *ipt_map;
+  realtype    a, b, rtemp, c, s, beta, onembeta, *cvals, *R, *gamma;
+  N_Vector    fv, vtemp, gold, fold, *df, *dg, *Q, *Xvecs;
+  booleantype damping;
 
   /* local shortcut variables */
   vtemp   = x;    /* use result as temporary vector */
@@ -425,6 +446,8 @@ static int AndersonAccelerate(SUNNonlinearSolver NLS, N_Vector gval,
   R       = FP_CONTENT(NLS)->R;
   gamma   = FP_CONTENT(NLS)->gamma;
   fv      = FP_CONTENT(NLS)->delta;
+  damping = FP_CONTENT(NLS)->damping;
+  beta    = FP_CONTENT(NLS)->beta;
 
   /* reset ipt_map, i_pt */
   for (i = 0; i < maa; i++)  ipt_map[i]=0;
@@ -479,7 +502,7 @@ static int AndersonAccelerate(SUNNonlinearSolver NLS, N_Vector gval,
       c = a / rtemp;
       s = b / rtemp;
       R[(i+1)*maa + i] = rtemp;
-      R[(i+1)*maa + i+1] = 0.0;
+      R[(i+1)*maa + i+1] = ZERO;
       if (i < maa-1) {
         for (j = i+2; j < maa; j++) {
           a = R[j*maa + i];
@@ -539,6 +562,19 @@ static int AndersonAccelerate(SUNNonlinearSolver NLS, N_Vector gval,
     nvec += 1;
   }
 
+  /* if enabled, apply damping */
+  if (damping) {
+    onembeta = (ONE - beta);
+    cvals[nvec] = -onembeta;
+    Xvecs[nvec] = fv;
+    nvec += 1;
+    for (i = lAA - 1; i > -1; i--) {
+      cvals[nvec] = onembeta * gamma[i];
+      Xvecs[nvec] = df[ipt_map[i]];
+      nvec += 1;
+    }
+  }
+
   /* update solution */
   retval = N_VLinearCombination(nvec, cvals, Xvecs, x);
   if (retval != 0)  return(SUN_NLS_VECTOROP_ERR);
@@ -582,7 +618,7 @@ static int AllocateContent(SUNNonlinearSolver NLS, N_Vector y)
     if (FP_CONTENT(NLS)->gamma == NULL) {
       FreeContent(NLS); return(SUN_NLS_MEM_FAIL); }
 
-    FP_CONTENT(NLS)->cvals = (realtype *) malloc((m+1) * sizeof(realtype));
+    FP_CONTENT(NLS)->cvals = (realtype *) malloc(2*(m+1) * sizeof(realtype));
     if (FP_CONTENT(NLS)->cvals == NULL) {
       FreeContent(NLS); return(SUN_NLS_MEM_FAIL); }
 
@@ -598,7 +634,7 @@ static int AllocateContent(SUNNonlinearSolver NLS, N_Vector y)
     if (FP_CONTENT(NLS)->q == NULL) {
       FreeContent(NLS); return(SUN_NLS_MEM_FAIL); }
 
-    FP_CONTENT(NLS)->Xvecs = (N_Vector *) malloc((m+1) * sizeof(N_Vector));
+    FP_CONTENT(NLS)->Xvecs = (N_Vector *) malloc(2*(m+1) * sizeof(N_Vector));
     if (FP_CONTENT(NLS)->Xvecs == NULL) {
       FreeContent(NLS); return(SUN_NLS_MEM_FAIL); }
   }
